module GhciExt
  (cmdHlp
  ,setOutToTmpFile
  ,restoreOutGetTmpFileContent
  ,restoreOutGetTmpFileName
  ,executeGrep
  ,extractArgUpToDelimiter
  ,parseStrArgs
  ,getSetStopCode
  ,getStopCode
  ,getLocateCode
  ,getSearchCode
  ,getRepeatCode
  ,highlightIf
  )
where

import GHC.IO.Handle
import System.IO
import System.Directory
import System.Process
import System.Exit
import Control.Monad
import Data.Maybe
import Data.List
import Data.Char
import Text.Regex


-- | Command help message handler.
cmdHlp :: ( String -> IO String ) -> String -> String -> String -> IO String
cmdHlp _ msg longMsg "--help" =
  let fullMsg = if null longMsg then msg++"\n" else msg ++ ('\n':longMsg) in
  return $ "putStr "++show fullMsg
cmdHlp _ msg _ "-h" = return $ "putStrLn "++show msg
cmdHlp action _ _ args = action args


-- | Redirects stadnard output to a temporary file.
setOutToTmpFile :: IO ( FilePath, Handle )
setOutToTmpFile = do
  ( tmpFileName, tmpFileHnd ) <- getTemporaryDirectory >>= flip openTempFile "ghciExt"
  orgStdOut <- hDuplicate stdout
  hDuplicateTo tmpFileHnd stdout
  hClose tmpFileHnd
  return ( tmpFileName, orgStdOut )

-- | Redirects stadnard output back from temporary file returning the file content.
restoreOutGetTmpFileContent :: ( FilePath, Handle ) -> IO String
restoreOutGetTmpFileContent ( tmpFileName, orgStdOut  ) = do
  hDuplicateTo orgStdOut stdout
  tmpFileText <- ( readFile tmpFileName >>= \t -> length t `seq` return t )
  removeFile tmpFileName
  return tmpFileText


-- | Redirects stadnard output back from temporary file returning the file name.
restoreOutGetTmpFileName :: ( FilePath, Handle ) -> IO String
restoreOutGetTmpFileName ( tmpFileName, orgStdOut  ) = do
  hDuplicateTo orgStdOut stdout
  return tmpFileName


-- | Executes OS grep on fileName merging :browse indented lines first.
executeGrep :: String -> String -> String -> IO ExitCode
executeGrep grepArgs cmd fileName = do
  when (isPrefixOf ":browse" cmd) mergeBrowseLinesInFile
  rv <- system $"grep "++grepArgs++" "++fileName
  removeFile fileName
  return rv

  where
    mergeBrowseLinesInFile = do
      fileText <- ( readFile fileName >>= \t -> length t `seq` return t )
      let mergedText = unlines $mergeBrowseLines $lines fileText
      writeFile fileName mergedText

    mergeBrowseLines []  = []
    mergeBrowseLines (ln0:ln1:lns)
      | indentSize ln1 > indentSize ln0  =
          let trimmedLn1 = ' ' : dropWhile isSpace ln1 in
          mergeBrowseLines ( (ln0++trimmedLn1) : lns )
      where
        indentSize ln = length $takeWhile isSpace ln
    mergeBrowseLines (ln:lns)  = ln : mergeBrowseLines lns


-- | Extracts argument up to specified delimiter.
-- Any delimiter character in shell-like quotations is not the considered to be
-- the argument separator which splits the string to the argument and the rest.
-- This is shell like (it does not skip any haskell end line comments).
extractArgUpToDelimiter :: Char -> String -> [String]
extractArgUpToDelimiter d argsStr = extractUpToDelimiter argsStr
  where
    extractUpToDelimiter str =
      let (prefix, suffix) = span isNotDelimOrStrStart str in
      if null suffix then [prefix] else
      let hdSuffix = head suffix in
      if d == hdSuffix then
        brushResultArgs (dropWhile isSpace) (trimRight prefix) (tail suffix) else
      let (string', prefixTail) = extractString hdSuffix (tail suffix) "" in
      let string = hdSuffix:string' in
      case extractUpToDelimiter prefixTail of
        [] -> [prefix++string]
        [restOfPrefix] -> [prefix++string++restOfPrefix]
        [restOfPrefix,realSuffix] -> [prefix++string++restOfPrefix, realSuffix]
        _ -> error "INTERNAL: extractUpToDelimiter returned more than two element list"

    extractString sd [] sRv = (reverse (sd:sRv), "")  -- add missing final string delimiter
    extractString sd ['\\'] sRv = (reverse (sd:sRv), "") -- drop unfinished escape
    extractString sd (hd:tl) sRv |sd==hd = (reverse (sd:sRv), tl) -- final string delimiter
    extractString sd ('\\':x:tl) sRv = extractString sd tl (x:'\\':sRv) -- escape
    extractString sd (hd:tl) sRv = extractString sd tl (hd:sRv) -- normal character

    trimRight s = reverse $dropWhile isSpace $reverse s

    isNotDelimOrStrStart x = x/=d && x/='\'' && x/='"'


-- | Extracts one argument from the string of arguments.
parseStrArg :: (String->String) -> String -> [String]
parseStrArg _ [] = []
parseStrArg adjustRestFn argStr@('"':_) =
  case reads argStr of
    ( (firstArg,rest) : _ ) -> brushResultArgs adjustRestFn firstArg rest
    _ -> [] -- string literal error; ignore the buggy part
parseStrArg adjustRestFn ('@':d:argStr) =
  case break (d==) argStr of
    ( firstArg, _:rest ) -> brushResultArgs adjustRestFn firstArg rest
    _ -> [ argStr ] -- we have only one argument actually
parseStrArg adjustRestFn argStr =
  case break isSpace argStr of
    ( firstArg, _:rest ) -> brushResultArgs adjustRestFn firstArg rest
    _ -> [ argStr ] -- we have only one argument actually


brushResultArgs :: (String->String) -> String -> String -> [String]
brushResultArgs adjustRestFn firstArg restOfArgs =
  case adjustRestFn restOfArgs of
    [] -> [firstArg]
    rest -> [firstArg, rest]


dropEndLineComment :: String -> String
dropEndLineComment [] = []
dropEndLineComment (c:cs) | isSpace c = dropEndLineComment cs
dropEndLineComment ('-':'-':_) = []
dropEndLineComment rest@(hd:_)
  | hd=='"'||hd=='@'  =  intercalate " " $parseStrArg dropEndLineComment rest
dropEndLineComment rest = dropCommentAfterUnquotedText rest
 where
  dropCommentAfterUnquotedText [] = []
  dropCommentAfterUnquotedText (x:xs) =
    if isSpace x && isDashDashAny xs then [] else
    x : dropCommentAfterUnquotedText xs

  isDashDashAny xs =
    case dropWhile isSpace xs of
      ('-':'-':_) -> True
      _ -> False


-- | Extracts given number of string arguments from the string of arguments.
-- Argument parsing drops any end-line comment following the last argument delimiter.
-- The argument can be:
-- + a Haskell string literal (the delimiter is the quotation mark)
-- + a text starting with @x - in such a case it is finished with the
--   next occurrence of x (x is the delimiter and represents any
--   character user selects)
-- + unquoted - it finishes with a whitespace (which is the delimiter)
parseStrArgs :: Int -> String -> [String]
parseStrArgs count argsStr =
  if count <= 0 then [ dropEndLineComment argsStr ] else
  if count == 1 then parseStrArg dropEndLineComment argsStr else
  case parseStrArg (dropWhile isSpace) argsStr of
    [] -> []
    [arg] -> [arg]
    [arg,rest] -> arg : parseStrArgs (count-1) rest
    _ -> error "INTERNAL: parseStrArg returned more than two element list"


-- | Generates code to setup stops.
-- The code will execute userCode when userCodeCheck is true and will stop at
-- the breakpoint when stopCheck is true.
-- * breakpointText is the output of :break command.
-- * userCode and stopCheck get identifiers ġė_BpId and ġė_HitCnt
--   macro expanded to breakpint identifier and hit count respectively
getSetStopCode :: String -> String -> String -> String -> String
getSetStopCode breakpointMsg userCodeCheck userCode stopCheck =
  case words breakpointMsg of
    "Breakpoint" : bpId : "activated" : _ ->
      let mergeLines = map (\c -> if isSpace c then ' ' else c) in
      let userCodeEx = macroExpand bpId userCode in
      let stopCheckEx = macroExpand bpId $ mergeLines stopCheck in
      let{ breakPointCode = "GhciExt.getStopCode "++bpId++" "++
        show (mergeLines userCodeCheck)++" "++show userCodeEx++" "++show stopCheckEx }in
      unlines
        [":def ġė_HitCnt"++bpId++" const$return \"let ġė_HitCnt"++bpId++"=0\""
        ,":set stop "++bpId++" :cmd return$"++breakPointCode ]
    _ -> "" -- break point creation failed (do not setup any stops)

  where
    macroExpand bpId text =
      let getReplaceRegexText s = "(^|[^[:alnum:]'_])"++s++"([^[:alnum:]'_]|$)" in
      let bpIdRegex = mkRegex $getReplaceRegexText "ġė_BpId" in
      let rv' = subRegex bpIdRegex text $ "\\1("++bpId++")\\2" in
      let hitCntRegex = mkRegex $getReplaceRegexText "ġė_HitCnt" in
      subRegex hitCntRegex rv' $ "\\1ġė_HitCnt"++bpId++"\\2"


-- | Generates breakpoint specific code to run when the breakpoint is hit.
getStopCode :: Int -> String -> String -> String -> String
getStopCode bpId userCodeCheck userCode stopCheck =
  showString ":init\n" $
  if null userCodeCheck then userCodeAndStopCheck else
  showString ":cmd return$ if " $ showString userCodeCheck $
    showString " then " $ shows userCodeAndStopCheck $
    showString " else "$ show continueCode

  where
    userCodeAndStopCheck =
      showString incBpHitCode $
      showString userCode $ showChar '\n' $
      showString ":cmd return$ if " $ showString stopCheck $
        showString " then " $ shows bpInfoCode $
        showString " else " $ show continueCode

    incBpHitCode = unlines
      [showString ":ġė_HitCnt"$ show bpId
      ,showString ":undef ġė_HitCnt" $ show bpId
      ,showString ":def ġė_HitCnt" $ shows bpId $ showString " const$ return$ " $
       showString  "\"let ġė_HitCnt" $ shows bpId $ showString "=\"++show(1+ġė_HitCnt" $ shows bpId ")"
      ,showString ":ġė_HitCnt" $ show bpId ]

    bpInfoCode =
      showString ":list\n" $
      showString "putStrLn $\"Stopped at breakpoint " $ shows bpId $
        showString "; hit count is \"++show ġė_HitCnt" $ show bpId

    continueCode = ":continue"


-- | Generates the code for :locate command.
getLocateCode :: String -> String -> String
getLocateCode breakpointMsg cmd =
  case parseBreakpointMsg of
    Just ( bpId, location, eraseBp ) -> unlines
            ["putStr \"Executing \""
            ,"putStr "++(show $show cmd)
            ,"putStr \" untill \""
            ,"putStr "++(show $show location)
            ,"putStrLn \" is hit.\""
            ,if eraseBp then ":delete "++bpId else ""
            ,":ġė_Search "++show(" at "++location++"\n")++"\"\""++cmd ] ;
    _ -> "putStrLn \"Cannot recognize the location.\""

  where
    parseBreakpointMsg =
      case words breakpointMsg of
        "Breakpoint" : bpId : "activated" : "at" : _ ->
          Just ( bpId, getLocation breakpointMsg, True )
        "Breakpoint" : bpId : "was" : "already" : "set" : "at" : _ ->
          Just ( bpId, getLocation breakpointMsg, False )
        _ -> Nothing

    getLocation [] = ""
    getLocation msg@(_:msgTl)
      | isPrefixOf " at " msg = takeWhile ('\n'/=) . drop 3 $ msgTl
      | otherwise = getLocation msgTl


-- | Handles unexpected results from stepping or history browsing.
handleUnknownStepResult :: String -> String -> String
handleUnknownStepResult text knowStepResultRv =
  if isPrefixOf "Logged breakpoint at " text  ||  isPrefixOf "Stopped at " text
    then knowStepResultRv
    else "putStr "++show text


-- | Generates the looping code for :find and :locate commands.
getSearchCode :: String -> String -> String -> String -> String
getSearchCode text pattern varName command = handleUnknownStepResult text knowStepResultRv
 where
  knowStepResultRv =
    if isInfixOf pattern text then
      (if null varName then "" else ":sprint "++varName++"\n") ++ ":list\n:init" else
    ":ġė_Search "++show pattern++show varName++command


-- | Generates the looping code for :repeat command.
getRepeatCode :: String -> String -> String -> String -> String
getRepeatCode text result expression command = handleUnknownStepResult text knowStepResultRv
 where
  knowStepResultRv =
    if filter (not . isSpace) result == "True" then ":list\n:init" else
    ":ġė_Repeat "++show expression++command


-- | Wrap string in a highlighting ansi escape sequence based on the regex match result.
highlightIf :: Bool -> String -> String -> String
highlightIf highlightMatch regexStr str =
  if highlightMatch /= isJust (matchRegex (mkRegex regexStr) str) then str else
  "\ESC[1;31m"++str++"\ESC[0m"

